
NAME = 'peony'

# 系统返回码
# 命令字不合法
RET_INVALID_CMD = -10000
# 系统内部异常
RET_INTERNAL = -10001
# admin用户验证失败
RET_ADMIN_AUTH_FAIL = -20000
# master连接未连接
RET_MASTER_NOT_CONNECTED = -21000


# 管理员命令
# 获取运行状态统计
CMD_ADMIN_SERVER_STAT = 20000

# 停止整个server
CMD_ADMIN_STOP = 21003


DEFAULT_CONFIG = {
    'HOST': '127.0.0.1',
    'PORT': 9250,

    # 启动的worker_id闭区间。
    'WORKER_ID_RANGE': None,

    # 通过task路由worker_id:
    #    def worker_router(task):
    #        return worker_id
    'WORKER_ROUTER': lambda task: 1,

    'DEBUG': False,

    # box class
    'BOX_CLASS': 'netkit.box.Box',

    # master class
    'MASTER_CLASS': 'peony.master.Master',

    # proxy class
    'PROXY_CLASS': 'peony.proxy.Proxy',

    # worker class
    'WORKER_CLASS': 'peony.worker.Worker',

    'NAME': NAME,

    # 每个worker的消息队列的最大长度, <=0 代表无限
    'TASK_QUEUE_MAX_SIZE': -1,

    # backlog
    'BACKLOG': 256,

    # 客户端连接超时
    'CLIENT_TIMEOUT': None,

    'STOP_TIMEOUT': None,

    # 处理task超时(秒). 超过后会打印fatal日志. None 代表永不超时
    'WORK_TIMEOUT': None,

    # 管理员，可以连接proxy获取数据
    # 管理员访问地址: 'admin.sock' or ('127.0.0.1', 9910)
    'ADMIN_ADDRESS': 'admin.sock',
    'ADMIN_USERNAME': None,
    'ADMIN_PASSWORD': None,
}



import sys
from collections import Counter

from .share.config import Config, ConfigAttribute
from .share.utils import import_module_or_string
from .share.log import logger
from .share import constants
from .share.mixins import RoutesMixin, AppEventsMixin


class Peony(RoutesMixin, AppEventsMixin):

    ############################## configurable begin ##############################
    name = ConfigAttribute('NAME')

    host = ConfigAttribute('HOST')
    port = ConfigAttribute('PORT')

    worker_id_range = ConfigAttribute('WORKER_ID_RANGE')

    worker_router = ConfigAttribute('WORKER_ROUTER')

    box_class = ConfigAttribute('BOX_CLASS',
                                get_converter=import_module_or_string)
    master_class = ConfigAttribute('MASTER_CLASS',
                                   get_converter=import_module_or_string)
    proxy_class = ConfigAttribute('PROXY_CLASS',
                                  get_converter=import_module_or_string)
    worker_class = ConfigAttribute('WORKER_CLASS',
                                   get_converter=import_module_or_string)
    debug = ConfigAttribute('DEBUG')

    task_queue_max_size = ConfigAttribute('TASK_QUEUE_MAX_SIZE')
    backlog = ConfigAttribute('BACKLOG')
    client_timeout = ConfigAttribute('CLIENT_TIMEOUT')
    stop_timeout = ConfigAttribute('STOP_TIMEOUT')
    work_timeout = ConfigAttribute('WORK_TIMEOUT')

    ############################## configurable end   ##############################

    config = None

    master = None
    blueprints = None

    def __init__(self):
        RoutesMixin.__init__(self)
        AppEventsMixin.__init__(self)

        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.blueprints = list()

    def register_blueprint(self, blueprint):
        blueprint.register_to_app(self)

    def run(self, host=None, port=None, debug=None):
        self._validate_cmds()

        if host is not None:
            self.config.update({
                'HOST': host,
            })

        if port is not None:
            self.config.update({
                'PORT': port,
            })

        if debug is not None:
            self.debug = debug

        # 在这里才创建master，从而保证config变量已经赋值了
        self.master = self.master_class(self)
        self.master.run()

    def make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s:%s %s] %s' % (
            constants.NAME,
            subtitle,
            self.name,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def _validate_cmds(self):
        """
        确保 cmd 没有重复
        :return:
        """

        cmd_list = list(self.rule_map.keys())

        for bp in self.blueprints:
            cmd_list.extend(list(bp.rule_map.keys()))

        duplicate_cmds = list((Counter(cmd_list) - Counter(set(cmd_list))).keys())

        assert not duplicate_cmds, 'duplicate cmds: %s' % duplicate_cmds

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.name
        )





"""
封装通过worker_id来访问数据的功能
"""

import multiprocessing
import queue
from ..share.log import logger


class TaskQueueManager(object):
    """
    通过worker_id来区分的queue
    """

    max_size = None
    queue_dict = None

    def __init__(self, worker_id_range, max_size=-1):
        self.max_size = max_size

        self.queue_dict = dict()
        for worker_id in range(worker_id_range[0], worker_id_range[1] + 1):
            self.queue_dict[worker_id] = multiprocessing.Queue(self.max_size)

    def get_queue(self, worker_id):
        """
        获取对应的队列
        :param worker_id:
        :return:
        """
        return self.queue_dict.get(worker_id)

    def put(self, worker_id, item):
        """
        加入item
        如果成功返回True，如果失败返回False
        :param worker_id:
        :param item:
        :return:
        """
        try:
            self.queue_dict[worker_id].put_nowait(item)
            return True
        except queue.Full:
            logger.error(
                'put item fail, queue is full. worker_id: %s, max_size: %s',
                worker_id, self.max_size
            )
            return False
        except:
            logger.error('put item fail, exc occur. worker_id: %s', worker_id, exc_info=True)
            return False

    def put_to_all(self, item):
        """
        向所有队列发送
        :param item:
        :return:
        """
        for q in self.queue_dict.values():
            try:
                q.put_nowait(item)
            except:
                pass

    def get(self, worker_id):
        if self.empty(worker_id):
            return None
        else:
            return self.queue_dict[worker_id].get_nowait()

    def clear(self, worker_id):
        """
        清空
        :param worker_id:
        :return:
        """
        self.queue_dict.pop(worker_id, None)

    def clear_all(self):
        """
        清空所有
        :return:
        """
        self.queue_dict.clear()

    def empty(self, worker_id):
        """
        判断是否是空的
        :param worker_id:
        :return:
        """
        return self.queue_dict[worker_id].empty()

    def qsize(self, worker_id):
        """
        某个队列的size
        :param worker_id:
        :return:
        """
        return self.queue_dict[worker_id].qsize()



from ..share.log import logger


class Request(object):
    """
    请求
    """

    # 业务层需要访问的对象，消息内容
    box = None

    worker = None
    # 封装的task，外面不需要理解
    task = None
    is_valid = False
    blueprint = None
    route_rule = None
    # 是否中断处理，即不调用view_func，主要用在before_request中
    interrupted = False

    def __init__(self, worker, task):
        self.worker = worker
        self.task = task
        # 赋值
        self.is_valid = self._parse_raw_data()

    def _parse_raw_data(self):
        if not self.task.body:
            return True

        self.box = self.app.box_class()

        if self.box.unpack(self.task.body) > 0:
            self._parse_route_rule()
            return True
        else:
            logger.error('unpack fail. request: %s', self)
            return False

    def _parse_route_rule(self):
        if self.cmd is None:
            return

        route_rule = self.app.get_route_rule(self.cmd)
        if route_rule:
            # 在app层，直接返回
            self.route_rule = route_rule
            return

        for bp in self.app.blueprints:
            route_rule = bp.get_route_rule(self.cmd)
            if route_rule:
                self.blueprint = bp
                self.route_rule = route_rule
                break

    @property
    def app(self):
        return self.worker.app

    @property
    def client_ip(self):
        """
        客户端连接IP，外面不需要了解task
        :return:
        """
        return self.task.client_ip

    @property
    def cmd(self):
        try:
            return self.box.cmd
        except:
            return None

    @property
    def view_func(self):
        return self.route_rule['view_func'] if self.route_rule else None

    @property
    def endpoint(self):
        if not self.route_rule:
            return None

        bp_endpoint = self.route_rule['endpoint']

        return '.'.join([self.blueprint.name, bp_endpoint] if self.blueprint else [bp_endpoint])

    def interrupt(self):
        """
        中断处理
        """
        self.interrupted = True

    def __repr__(self):
        return '<%s cmd: %r, endpoint: %s, task: %r, worker: %s>' % (
            type(self).__name__, self.cmd, self.endpoint, self.task, self.worker
        )



import queue
import signal
import time
from pyglet.clock import Clock
import _thread
import setproctitle
from maple import constants as maple_constants

from ..share import constants
from ..share.log import logger
from .request import Request


class Worker:
    request_class = Request
    got_first_request = False

    # pipe 比 queue性能高一点，因为queue是在pipe的基础上加了锁
    # 但是最终还是得选择queue，因为pipe.recv是阻塞的，也就是每次poll完了只能读取一次
    # 这会导致clock的计算次数增加，没必要
    task_queue = None

    app = None
    worker_id = None

    enabled = True
    got_keyboard_interrupt = False

    # 业务可以通过clock来设置schedule
    clock = None

    # 工作进展
    work_progress = None

    def __init__(self, app, worker_id, task_queue):
        self.app = app
        self.worker_id = worker_id
        self.task_queue = task_queue

        self.clock = Clock()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            'worker:%s' % self.worker_id
        ))

        self._handle_signals()

        self._on_start()

        # daemon==True，主线程不需要等待网络线程结束
        _thread.start_new_thread(self._monitor_work_timeout, ())

        while self.enabled:
            try:
                self._handle()
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur. worker: %s', self, exc_info=True)

        self._on_stop()

    def _handle(self):
        """
        主循环，等待网络消息，并进行处理
        """
        # 允许idle
        sleep_time = self.clock.get_sleep_time(True)
        # sleep_time is None 说明没有需要处理的，直接永久阻塞等待即可
        # 否则就要使用指定时间阻塞等待
        task_list = []

        # 第一个get要block
        block = True
        while True:
            try:
                task_list.append(
                    self.task_queue.get(block=block, timeout=sleep_time)
                )
            except queue.Empty:
                # 说明已经没有了
                break
            except KeyboardInterrupt:
                # 原样抛出
                raise
            except:
                # 说明出问题了
                logger.error('exc occur. worker: %s', self, exc_info=True)
                break
            finally:
                block = False

        for task in task_list:
            if task is None:
                # 通过传入None来中断block
                continue

            request = self.request_class(self, task)
            # 设置task开始处理的时间和信息
            self.work_progress = dict(
                begin_time=time.time(),
                request=request,
            )

            self._handle_request(request)

            self.work_progress = None

        self.clock.tick(True)

    def _handle_request(self, request):
        """
        出现任何异常的时候，服务器不再主动关闭连接
        """
        if not request.is_valid:
            return False

        # 没有办法得到 client_created 事件，因为这个时候还没有标识信息
        if request.task.cmd == maple_constants.CMD_CLIENT_CLOSED:
            self.app.events.close_client(request)
            for bp in self.app.blueprints:
                bp.events.close_app_client(request)
            return True

        if not request.view_func:
            logger.info('cmd invalid. request: %s' % request)
            return False

        if not self.got_first_request:
            self.got_first_request = True
            self.app.events.before_first_request(request)
            for bp in self.app.blueprints:
                bp.events.before_app_first_request(request)

        self.app.events.before_request(request)
        for bp in self.app.blueprints:
            bp.events.before_app_request(request)
        if request.blueprint:
            request.blueprint.events.before_request(request)

        if request.interrupted:
            # 业务要求中断
            return True

        view_func_exc = None

        try:
            request.view_func(request)
        except Exception as e:
            logger.error('view_func raise exception. e: %s, request: %s', e, request, exc_info=True)
            view_func_exc = e

        if request.blueprint:
            request.blueprint.events.after_request(request, view_func_exc)
        for bp in self.app.blueprints:
            bp.events.after_app_request(request, view_func_exc)
        self.app.events.after_request(request, view_func_exc)

        return True

    def _monitor_work_timeout(self):
        """
        监控task的耗时
        :return:
        """

        while True:
            time.sleep(1)

            work_progress = self.work_progress
            if work_progress:
                past_time = time.time() - work_progress['begin_time']
                if self.app.work_timeout is not None and past_time > self.app.work_timeout:
                    # 说明worker的处理时间已经太长了
                    logger.fatal('work timeout: %s / %s, request: %s',
                                 past_time, self.app.work_timeout, work_progress['request'])
                    # 不能这么做啊，直接把其他room也杀掉了，就打印一下告警算了
                    # 强制从子线程退出整个进程
                    # os._exit(-1)

    def push_task(self, task):
        """
        加入task，不阻塞
        return:
            True: 成功
            False: 失败
        """
        try:
            self.task_queue.put_nowait(task)
            return True
        except:
            logger.error('exc occur. worker: %s, task: %s', self, task, exc_info=True)
            return False

    def _on_start(self):
        """
        当worker启动后
        可继承实现
        :return:
        """
        self.app.events.start_worker(self)
        for bp in self.app.blueprints:
            bp.events.start_app_worker(self)

    def _on_stop(self):
        """
        当worker停止前
        可继承实现
        :return:
        """
        for bp in self.app.blueprints:
            bp.events.stop_app_worker(self)
        self.app.events.stop_worker(self)

    def _handle_signals(self):
        def stop_handler(signum, frame):
            self.enabled = False
            if not self.got_keyboard_interrupt:
                self.got_keyboard_interrupt = True
                raise KeyboardInterrupt

        def safe_stop_handler(signum, frame):
            self.enabled = False

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, stop_handler)
        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    def get_fps(self):
        return self.clock.get_fps()

    def schedule(self, func, *args, **kwargs):
        self.clock.schedule(func, *args, **kwargs)

    def schedule_interval(self, func, interval, *args, **kwargs):
        self.clock.schedule_interval(func, interval, *args, **kwargs)

    def schedule_interval_soft(self, func, interval, *args, **kwargs):
        self.clock.schedule_interval_soft(func, interval, *args, **kwargs)

    def schedule_once(self, func, delay, *args, **kwargs):
        self.clock.schedule_once(func, delay, *args, **kwargs)

    def unschedule(self, func):
        self.clock.unschedule(func)

    def unschedule_all(self):
        self.clock = Clock()

    def __repr__(self):
        return '<%s name: %s, worker_id: %r>' % (
            type(self).__name__, self.app.name, self.worker_id
        )




from collections import defaultdict


class StatCounter(object):

    """
    统计计算类
    """

    # 客户端连接数
    clients = 0
    # 客户端请求数
    client_req = 0
    # 客户端回应数
    client_rsp = 0
    # worker请求数
    worker_req_counter = None
    # 丢弃的任务
    discard_tasks_counter = None

    def __init__(self):
        self.worker_req_counter = defaultdict(int)
        self.discard_tasks_counter = defaultdict(int)

    def add_worker_req(self, worker_id):
        self.worker_req_counter[worker_id] += 1

    def add_discard_task(self, worker_id):
        self.discard_tasks_counter[worker_id] += 1




import multiprocessing
import signal
import time
import threading
import _thread
import setproctitle
from .task_queue_manager import TaskQueueManager
from ..share.utils import safe_call
from ..share.log import logger


class Master:

    app = None

    proxy = None
    worker_processes = None

    enabled = True

    task_queue_manager = None

    def __init__(self, app):
        self.app = app
        self.worker_processes = list()
        self.task_queue_manager = TaskQueueManager(
            self.app.worker_id_range, self.app.task_queue_max_size
        )

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            'master:%s-%s' % (self.app.worker_id_range[0], self.app.worker_id_range[1])
        ))
        self._handle_signals()

        self._spawn_workers()

        thread_list = [
            # 监视workers
            threading.Thread(target=self._monitor_workers),
            # proxy
            threading.Thread(target=self._run_proxy)
        ]

        for t in thread_list:
            # 不可以独自退出
            t.daemon = False
            t.start()

        for t in thread_list:
            t.join()

    def _run_proxy(self):
        self.proxy = self.app.proxy_class(self.app)
        self.proxy.run()

    def push_task(self, worker_id, task):
        """
        指定worker的task
        :param worker_id:
        :param task:
        :return:
        """

        return self.task_queue_manager.put(worker_id, task)

    def _create_worker(self, worker_id, task_queue):
        process = multiprocessing.Process(
            target=self.app.worker_class(self.app, worker_id, task_queue).run,
        )
        # master不可以独自退出
        process.daemon = False
        # 标记worker_id, conn
        process.init_params = dict(
            worker_id=worker_id,
            task_queue=task_queue
        )

        process.start()

        return process

    def _spawn_workers(self):
        """
        监控进程
        :return:
        """
        for worker_id in range(self.app.worker_id_range[0], self.app.worker_id_range[1] + 1):
            self.worker_processes.append(self._create_worker(
                worker_id,
                self.task_queue_manager.get_queue(worker_id),
            ))

    def _monitor_workers(self):
        while True:
            for idx, process in enumerate(self.worker_processes):
                if process and not process.is_alive():
                    self.worker_processes[idx] = None

                    if self.enabled:
                        # 需要重启启动process
                        self.worker_processes[idx] = self._create_worker(**process.init_params)

            if not self.enabled and not any(self.worker_processes):
                # 可以彻底停掉了
                break

            time.sleep(0.1)

    def stop_all(self):
        """
        等所有子进程结束，父进程也退出
        """

        def kill_processes_later(processes, timeout):
            """
            等待一段时间后杀死所有进程
            :param processes:
            :param timeout:
            :return:
            """

            def _kill_processes():
                # 等待一段时间
                time.sleep(timeout)

                for p in processes:
                    if p and p.is_alive():
                        # 说明进程还活着
                        p.kill()

            _thread.start_new_thread(_kill_processes, ())

        self.enabled = False

        # 为了中断worker的阻塞等待
        self.task_queue_manager.put_to_all(None)

        # 一定要这样，否则后面kill的时候可能会kill错
        processes = self.worker_processes[:]

        # 不能只靠task is None来处理，万一现在队列满了呢？

        # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
        # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
        # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
        for p in processes:
            if p:
                p.terminate()

        if self.app.stop_timeout is not None:
            kill_processes_later(processes, self.app.stop_timeout)

        # 关闭proxy
        self.proxy.stop()

    def _handle_signals(self):

        def stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self.stop_all()

        # INT为强制结束
        signal.signal(signal.SIGINT, stop_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.app.name
        )




import signal
import os
# linux 默认就是epoll
from twisted.internet import reactor

from .connection.client_tcp_connection import ClientConnectionFactory as ClientTCPConnectionFactory
from .connection.admin_connection import AdminConnectionFactory
from .stat_counter import StatCounter
from ..share import constants
from ..share.log import logger
from ..share.utils import safe_call


class Proxy(object):
    """
    proxy相关
    """

    client_tcp_connection_factory_class = ClientTCPConnectionFactory

    admin_connection_factory_class = AdminConnectionFactory

    app = None

    # 统计
    stat_counter = None

    def __init__(self, app):
        """
        构造函数
        :return:
        """
        self.app = app
        self.stat_counter = StatCounter()

    def run(self):
        # 启动对外监听
        host_list = self.app.host
        if isinstance(host_list, str):
            # 说明是字符串，要转化成数组
            host_list = (host_list,)

        for host in host_list:
            reactor.listenTCP(self.app.port, self.client_tcp_connection_factory_class(self),
                              backlog=self.app.backlog, interface=host)

        # 启动admin服务
        admin_address = self.app.config['ADMIN_ADDRESS']

        if admin_address:
            if isinstance(admin_address, (list, tuple)):
                # 说明是网络协议
                reactor.listenTCP(admin_address[1], self.admin_connection_factory_class(self),
                                  interface=admin_address[0])
            elif isinstance(admin_address, str):
                # 说明是文件
                if os.path.exists(admin_address):
                    os.remove(admin_address)

                # 创建目录
                admin_directory = os.path.dirname(admin_address)
                if admin_directory and not os.path.exists(admin_directory):
                    os.makedirs(admin_directory)

                reactor.listenUNIX(admin_address, self.admin_connection_factory_class(self))
            else:
                logger.error('invalid admin address. proxy: %s, admin_address: %s', self, admin_address)

        try:
            reactor.run(installSignalHandlers=False)
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur. proxy: %s', self, exc_info=True)

    def stop(self):
        """
        停止
        :return:
        """
        try:
            # 直接调用stop是关闭不了的
            reactor.callFromThread(reactor.stop)
        except:
            logger.error('exc occur.', exc_info=True)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.app.name
        )
